@@ -2,18 +2,18 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 #
-# Copyright (c) 2014 Etherios, Inc. All rights reserved.
-# Etherios, Inc. is a Division of Digi International.
+# Copyright (c) 2015-2018 Digi International Inc.
 
 r"""Module providing classes for interacting with device cloud data streams"""
-
+import json
 import logging
 import datetime
 
 import six
 from devicecloud.apibase import APIBase
 from devicecloud import DeviceCloudException, DeviceCloudHttpException
-from devicecloud.util import conditional_write, to_none_or_dt, validate_type, isoformat
+from devicecloud.util import conditional_write, to_none_or_dt, validate_type, isoformat, \
+    dc_utc_timestamp_to_dt
 from six import StringIO
 
 
@@ -25,6 +25,7 @@ STREAM_TYPE_FLOAT = "FLOAT"
 STREAM_TYPE_DOUBLE = "DOUBLE"
 STREAM_TYPE_STRING = "STRING"
 STREAM_TYPE_BINARY = "BINARY"
+STREAM_TYPE_JSON = "JSON"
 STREAM_TYPE_UNKNOWN = "UNKNOWN"
 
 ROLLUP_INTERVAL_HALF = "half"
@@ -53,6 +54,7 @@ DSTREAM_TYPE_MAP = {
     STREAM_TYPE_STRING: (str, str),
     STREAM_TYPE_BINARY: (str, str),
     STREAM_TYPE_UNKNOWN: (str, str),
+    STREAM_TYPE_JSON: (json.loads, json.dumps)
 }
 
 
@@ -61,6 +63,34 @@ ONE_DAY = 86400  # in seconds
 logger = logging.getLogger("devicecloud.streams")
 
 
+def _get_encoder_method(stream_type):
+    """A function to get the python type to device cloud type converter function.
+
+    :param stream_type: The streams data type
+    :return: A function that when called with the python object will return the serializable
+    type for sending to the cloud. If there is no function for the given type, or the `stream_type`
+    is `None` the returned function will simply return the object unchanged.
+    """
+    if stream_type is not None:
+        return DSTREAM_TYPE_MAP.get(stream_type.upper(), (lambda x: x, lambda x: x))[1]
+    else:
+        return lambda x: x
+
+
+def _get_decoder_method(stream_type):
+    """ A function to get Device Cloud type to python type converter function.
+
+    :param stream_type: The streams data type
+    :return: A function that when called with Device Cloud object will return the python
+    native type. If there is no function for the given type, or the `stream_type` is `None`
+    the returned function will simply return the object unchanged.
+    """
+    if stream_type is not None:
+        return DSTREAM_TYPE_MAP.get(stream_type.upper(), (lambda x: x, lambda x: x))[0]
+    else:
+        return lambda x: x
+
+
 class StreamException(DeviceCloudException):
     """Base class for stream related exceptions"""
 
@@ -69,6 +99,10 @@ class NoSuchStreamException(StreamException):
     """Failure to find a stream based on a given id"""
 
 
+class InvalidRollupDatatype(StreamException):
+    """Roll-up's are only valid on numerical data types"""
+
+
 class StreamsAPI(APIBase):
     """Provide interface for interacting with device cloud streams API
 
@@ -79,11 +113,15 @@ class StreamsAPI(APIBase):
     def __init__(self, *args, **kwargs):
         APIBase.__init__(self, *args, **kwargs)
 
-    def _get_streams(self):
+    def _get_streams(self, uri_suffix=None):
         """Clear and update internal cache of stream objects"""
         # TODO: handle paging, perhaps change this to be a generator
+        if uri_suffix is not None and not uri_suffix.startswith('/'):
+            uri_suffix = '/' + uri_suffix
+        elif uri_suffix is None:
+            uri_suffix = ""
         streams = {}
-        response = self._conn.get_json("/ws/DataStream")
+        response = self._conn.get_json("/ws/DataStream{}".format(uri_suffix))
         for stream_data in response["items"]:
             stream_id = stream_data["streamId"]
             stream = DataStream(self._conn, stream_id, stream_data)
@@ -92,12 +130,12 @@ class StreamsAPI(APIBase):
 
     def create_stream(self, stream_id, data_type, description=None, data_ttl=None,
                       rollup_ttl=None, units=None):
-        """Create a new data stream on the device cloud
+        """Create a new data stream on Device Cloud
 
-        This method will attempt to create a new data stream on the device cloud.
+        This method will attempt to create a new data stream on Device Cloud.
         This method will only succeed if the stream does not already exist.
 
-        :param str stream_id: The path/id of the stream being created on the device cloud.
+        :param str stream_id: The path/id of the stream being created on Device Cloud.
         :param str data_type: The type of this stream.  This must be in the set
             `{ INTEGER, LONG, FLOAT, DOUBLE, STRING, BINARY, UNKNOWN }`.  These values are
             available in constants like :attr:`~STREAM_TYPE_INTEGER`.
@@ -114,7 +152,7 @@ class StreamsAPI(APIBase):
             data_type = str(data_type).upper()
         if not data_type in (set([None, ]) | set(list(DSTREAM_TYPE_MAP.keys()))):
             raise ValueError("data_type %r is not valid" % data_type)
-        description = validate_type(stream_id, type(None), *six.string_types)
+        description = validate_type(description, type(None), *six.string_types)
         data_ttl = validate_type(data_ttl, type(None), *six.integer_types)
         rollup_ttl = validate_type(rollup_ttl, type(None), *six.integer_types)
         units = validate_type(units, type(None), *six.string_types)
@@ -134,14 +172,16 @@ class StreamsAPI(APIBase):
         stream = DataStream(self._conn, stream_id)
         return stream
 
-    def get_streams(self):
-        """Return the iterator over all streams present on the device cloud
+    def get_streams(self, stream_prefix=None):
+        """Return the iterator over streams preset on device cloud.
 
-        :return:  iterator over all :class:`.DataStream` instances on the device cloud
+        :param stream_prefix: An optional prefix to limit the iterator to; all streams are returned if it is not specified.
+
+        :return:  iterator over all :class:`.DataStream` instances on Device Cloud
 
         """
         # TODO: deal with paging.  We now return a generator, so the interface should look the same
-        return iter(self._get_streams().values())
+        return iter(self._get_streams(stream_prefix).values())
 
     def get_stream(self, stream_id):
         """Return a reference to a stream with the given ``stream_id``
@@ -151,7 +191,7 @@ class StreamsAPI(APIBase):
         stream exists, one can use :py:meth:`get_stream_if_exists` which will
         return None if the stream is not already created.
 
-        :param stream_id: The path of the stream on the device cloud
+        :param stream_id: The path of the stream on Device Cloud
         :raises TypeError: if the stream_id provided is the wrong type
         :raises ValueError: if the stream_id is not properly formed
         :return: datastream instance with the provided stream_id
@@ -166,7 +206,7 @@ class StreamsAPI(APIBase):
         This works similar to :py:meth:`get_stream` but will return None if the
         stream is not already created.
 
-        :param stream_id: The path of the stream on the device cloud
+        :param stream_id: The path of the stream on Device Cloud
         :raises TypeError: if the stream_id provided is the wrong type
         :raises ValueError: if the stream_id is not properly formed
         :return: :class:`.DataStream` instance with the provided stream_id
@@ -185,7 +225,7 @@ class StreamsAPI(APIBase):
         """Perform a bulk write (or set of writes) of a collection of data points
 
         This method takes a list (or other iterable) of datapoints and writes them
-        to the device cloud in an efficient manner, minimizing the number of HTTP
+        to Device Cloud in an efficient manner, minimizing the number of HTTP
         requests that need to be made.
 
         As this call is performed from outside the context of any particular stream,
@@ -208,14 +248,14 @@ class StreamsAPI(APIBase):
             dc.streams.bulk_write_datapoints(datapoints)
 
         Depending on the size of the list of datapoints provided, this method may
-        need to make multiple calls to the device cloud (in chunks of 250).
+        need to make multiple calls to Device Cloud (in chunks of 250).
 
-        :param list datapoints: a list of datapoints to be written to the device cloud
+        :param list datapoints: a list of datapoints to be written to Device Cloud
         :raises TypeError: if a list of datapoints is not provided
         :raises ValueError: if any of the provided data points do not have all required
             information (such as information about the stream)
         :raises DeviceCloudHttpException: in the case of an unexpected error in communicating
-            with the device cloud.
+            with Device Cloud.
 
         """
         datapoints = list(datapoints)  # effectively performs validation that we have the right type
@@ -248,7 +288,7 @@ class DataPoint(object):
 
     This class encapsulates the data required for both pushing data points to the
     device cloud as well as for storing and provding methods to access data from
-    streams that has been retrieved from the device cloud.
+    streams that has been retrieved from Device Cloud.
 
     """
 
@@ -257,11 +297,13 @@ class DataPoint(object):
         """Create a new DataPoint object from device cloud JSON data
 
         :param DataStream stream: The :class:`~DataStream` out of which this data is coming
-        :param dict json_data: Deserialized JSON data from the device cloud about this device
+        :param dict json_data: Deserialized JSON data from Device Cloud about this device
         :raises ValueError: if the data is malformed
         :return: (:class:`~DataPoint`) newly created :class:`~DataPoint`
 
         """
+        type_converter = _get_decoder_method(stream.get_data_type())
+        data = type_converter(json_data.get("data"))
         return cls(
             # these are actually properties of the stream, not the data point
             stream_id=stream.get_stream_id(),
@@ -269,7 +311,7 @@ class DataPoint(object):
             units=stream.get_units(),
 
             # and these are part of the data point itself
-            data=json_data.get("data"),
+            data=data,
             description=json_data.get("description"),
             timestamp=json_data.get("timestampISO"),
             server_timestamp=json_data.get("serverTimestampISO"),
@@ -278,6 +320,29 @@ class DataPoint(object):
             dp_id=json_data.get("id"),
         )
 
+    @classmethod
+    def from_rollup_json(cls, stream, json_data):
+        """Rollup json data from the server looks slightly different
+
+        :param DataStream stream: The :class:`~DataStream` out of which this data is coming
+        :param dict json_data: Deserialized JSON data from Device Cloud about this device
+        :raises ValueError: if the data is malformed
+        :return: (:class:`~DataPoint`) newly created :class:`~DataPoint`
+        """
+        dp = cls.from_json(stream, json_data)
+
+        # Special handling for timestamp
+        timestamp = isoformat(dc_utc_timestamp_to_dt(int(json_data.get("timestamp"))))
+
+        # Special handling for data, all rollup data is float type
+        type_converter = _get_decoder_method(stream.get_data_type())
+        data = type_converter(float(json_data.get("data")))
+
+        # Update the special fields
+        dp.set_timestamp(timestamp)
+        dp.set_data(data)
+        return dp
+
     def __init__(self, data, stream_id=None, description=None, timestamp=None,
                  quality=None, location=None, data_type=None, units=None, dp_id=None,
                  customer_id=None, server_timestamp=None):
@@ -346,12 +411,7 @@ class DataPoint(object):
 
     def get_data(self):
         """Get the actual data value associated with this data point"""
-        data = self._data
-        if self._data_type is not None:
-            type_converters = DSTREAM_TYPE_MAP.get(self._data_type.upper())
-            if type_converters:
-                data = type_converters[0](self._data)
-        return data
+        return self._data
 
     def set_data(self, data):
         """Set the data for this data point
@@ -415,7 +475,7 @@ class DataPoint(object):
     def set_quality(self, quality):
         """Set the quality for this sample
 
-        Quality is stored on the device cloud as a 32-bit integer, so the input
+        Quality is stored on Device Cloud as a 32-bit integer, so the input
         to this function should be either None, an integer, or a string that can
         be converted to an integer.
 
@@ -517,10 +577,14 @@ class DataPoint(object):
         set on this datapoint.  Values not set (e.g. quality) will be ommitted.
 
         """
+        type_converter = _get_encoder_method(self._data_type)
+        # Convert from python native to device cloud
+        encoded_data = type_converter(self._data)
+
         out = StringIO()
         out.write("<DataPoint>")
         out.write("<streamId>{}</streamId>".format(self.get_stream_id()))
-        out.write("<data>{}</data>".format(self.get_data()))
+        out.write("<data>{}</data>".format(encoded_data))
         conditional_write(out, "<description>{}</description>", self.get_description())
         if self.get_timestamp() is not None:
             out.write("<timestamp>{}</timestamp>".format(isoformat(self.get_timestamp())))
@@ -574,13 +638,13 @@ class DataStream(object):
             ))
 
     def _get_stream_metadata(self, use_cached):
-        """Retrieve metadata about this stream from the device cloud"""
+        """Retrieve metadata about this stream from Device Cloud"""
         if self._cached_data is None or not use_cached:
             try:
                 self._cached_data = self._conn.get_json("/ws/DataStream/%s" % self._stream_id)["items"][0]
             except DeviceCloudHttpException as http_exception:
                 if http_exception.response.status_code == 404:
-                    raise NoSuchStreamException("Stream with id %r has not been created", self._stream_id)
+                    raise NoSuchStreamException("Stream with id %r has not been created" % self._stream_id)
                 raise http_exception
         return self._cached_data
 
@@ -610,7 +674,7 @@ class DataStream(object):
         * BINARY - Data with this type map to a python string.
         * UNKNOWN - Data with this type map to a python string.
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :return: The data type of this stream as a string
         :rtype: str
@@ -626,7 +690,7 @@ class DataStream(object):
 
         Units are a user-defined field stored as a string
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :return: The unit of this stream as a string
         :rtype: str or None
@@ -637,7 +701,7 @@ class DataStream(object):
     def get_description(self, use_cached=True):
         """Get the description associated with this data stream
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :raises devicecloud.DeviceCloudHttpException: in the case of an unexpected http error
         :raises devicecloud.streams.NoSuchStreamException: if this stream has not yet been created
@@ -653,7 +717,7 @@ class DataStream(object):
         The dataTtl is the time to live (TTL) in seconds for data points stored in the data stream.
         A data point expires after the configured amount of time and is automatically deleted.
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :raises devicecloud.DeviceCloudHttpException: in the case of an unexpected http error
         :raises devicecloud.streams.NoSuchStreamException: if this stream has not yet been created
@@ -672,7 +736,7 @@ class DataStream(object):
         stored in the stream. A roll-up expires after the configured amount of time and is
         automatically deleted.
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :raises devicecloud.DeviceCloudHttpException: in the case of an unexpected http error
         :raises devicecloud.streams.NoSuchStreamException: if this stream has not yet been created
@@ -688,7 +752,7 @@ class DataStream(object):
 
         The current value is the last recorded data point for this stream.
 
-        :param bool use_cached: If False, the function will always request the latest from the device cloud.
+        :param bool use_cached: If False, the function will always request the latest from Device Cloud.
             If True, the device will not make a request if it already has cached data.
         :raises devicecloud.DeviceCloudHttpException: in the case of an unexpected http error
         :raises devicecloud.streams.NoSuchStreamException: if this stream has not yet been created
@@ -703,7 +767,7 @@ class DataStream(object):
             return None
 
     def delete(self):
-        """Delete this stream from the device cloud along with its history
+        """Delete this stream from Device Cloud along with its history
 
         This call will return None on success and raise an exception in the event of an error
         performing the deletion.
@@ -808,7 +872,7 @@ class DataStream(object):
 
         Values already set on the datapoint will not be overridden (except for path)
 
-        :param DataPoint datapoint: The :class:`.DataPoint` that should be written to the device cloud
+        :param DataPoint datapoint: The :class:`.DataPoint` that should be written to Device Cloud
 
         """
         if not isinstance(datapoint, DataPoint):
@@ -825,13 +889,13 @@ class DataStream(object):
         """Read one or more DataPoints from a stream
 
         .. warning::
-           The data points from the device cloud is a paged data set.  When iterating over the
+           The data points from Device Cloud is a paged data set.  When iterating over the
            result set there could be delays when we hit the end of a page.  If this is undesirable,
            the caller should collect all results into a data structure first before iterating over
            the result set.
 
         :param start_time: The start time for the window of data points to read.  None means
-            that we should start with the old data available.
+            that we should start with the oldest data available.
         :type start_time: :class:`datetime.datetime` or None
         :param end_time: The end time for the window of data points to read.  None means
             that we should include all points received until this point in time.
@@ -868,6 +932,23 @@ class DataStream(object):
         :returns: A generator object which one can iterate over the DataPoints read.
 
         """
+
+        is_rollup = False
+        if (rollup_interval is not None) or (rollup_method is not None):
+            is_rollup = True
+            numeric_types = [
+                STREAM_TYPE_INTEGER,
+                STREAM_TYPE_LONG,
+                STREAM_TYPE_FLOAT,
+                STREAM_TYPE_DOUBLE,
+                STREAM_TYPE_STRING,
+                STREAM_TYPE_BINARY,
+                STREAM_TYPE_UNKNOWN,
+            ]
+
+            if self.get_data_type(use_cached=True) not in numeric_types:
+                raise InvalidRollupDatatype('Rollups only support numerical DataPoints')
+
         # Validate function inputs
         start_time = to_none_or_dt(validate_type(start_time, datetime.datetime, type(None)))
         end_time = to_none_or_dt(validate_type(end_time, datetime.datetime, type(None)))
@@ -929,5 +1010,8 @@ class DataStream(object):
             result_size = int(result["resultSize"])  # how many are actually included here?
             query_parameters["pageCursor"] = result.get("pageCursor")  # will not be present if result set is empty
             for item_info in result.get("items", []):
-                data_point = DataPoint.from_json(self, item_info)
+                if is_rollup:
+                    data_point = DataPoint.from_rollup_json(self, item_info)
+                else:
+                    data_point = DataPoint.from_json(self, item_info)
                 yield data_point
